package cat.hack3.codingtests.marsrover.cartography;

import java.util.List;
import java.util.logging.Logger;

/**
 * Notice that the representation is the following:
 * Latitude is Y, starts from 1 on the top and goes down incrementally
 * Longitude is X, starts from 1 on the left and goes to right incrementally
 *     1    2    3
 * 1  1-1  1-2  1-3
 * 2  2-1  2-2  2-3
 * 3  3-1  3-2  3-3
 */
public class MarsMap {
    Logger logger = Logger.getLogger(this.getClass().getName());
    private static final int FIRST_POSITION_IN_MAP = 1;
    private static final int INCREMENT_UNIT = 1;

    private final int height;
    private final int width;
    private Coordinates currentPosition;
    private Direction currentMovementDirection;
    private final List<Coordinates> obstaclesLocalizations;

    private final MapIncrementalPositionResolver positionResolver;

    @SafeVarargs
    public MarsMap(int height, int width, Coordinates startingCoordinates, List<Coordinates>... obstaclesLocalizations) {
        this.height = height;
        this.width = width;
        currentPosition = startingCoordinates;
        this.obstaclesLocalizations = obstaclesLocalizations.length > 0 ? obstaclesLocalizations[0] : List.of();
        positionResolver = new MapIncrementalPositionResolver(FIRST_POSITION_IN_MAP, INCREMENT_UNIT, height, width);
    }

    public Coordinates getNextPositionTowards(Direction direction) {
        return switch (direction) {
            case NORTH -> getDecrementLatitude();
            case SOUTH -> getIncrementLatitude();
            case WEST -> getDecrementLongitude();
            case EAST -> getIncrementLongitude();
        };
    }

    private Coordinates getIncrementLatitude() {
        var latitude = currentPosition.latitude();
        long newLatitude = positionResolver.getIncremented(latitude);
        return currentPosition.ofUpdatedLatitude(newLatitude);
    }

    private Coordinates getDecrementLatitude() {
        var latitude = currentPosition.latitude();
        long newLatitude = positionResolver.getDecremented(latitude);
        return currentPosition.ofUpdatedLatitude(newLatitude);
    }

    private Coordinates getIncrementLongitude() {
        var longitude = currentPosition.longitude();
        long newLongitude = positionResolver.getIncremented(longitude);
        return currentPosition.ofUpdatedLongitude(newLongitude);
    }

    private Coordinates getDecrementLongitude() {
        var longitude = currentPosition.longitude();
        long newLongitude = positionResolver.getDecremented(longitude);
        return currentPosition.ofUpdatedLongitude(newLongitude);
    }

    public synchronized void updateLocation(Coordinates newPosition) {
        currentPosition = newPosition;
    }

    public boolean willCollideWithObstacle(Coordinates newCoordinates) {
        return obstaclesLocalizations.contains(newCoordinates);
    }

    public Coordinates getCurrentPosition() {
        return currentPosition;
    }
}
